import json
from contextlib import suppress
from functools import cached_property

import numpy as np
import pandas as pd
from django.core.cache import cache
from django.db.models import QuerySet
from rest_framework.mixins import ListModelMixin
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet

from wbcore.cache.mixins import CacheMixin
from wbcore.contrib.pandas import fields
from wbcore.contrib.pandas.filters import PandasDjangoFilterBackend, PandasOrderingFilter, PandasSearchFilter
from wbcore.contrib.pandas.metadata import PandasMetadata
from wbcore.metadata.mixins import WBCoreMetadataConfigViewMixin
from wbcore.utils.renderers import BrowsableAPIRendererWithoutForms
from wbcore.viewsets.encoders import PandasDataCustomRenderer
from wbcore.viewsets.mixins import DocumentationMixin, MessageMixin


class PandasMixin(CacheMixin):
    filter_backends = [
        PandasOrderingFilter,
        PandasSearchFilter,
        PandasDjangoFilterBackend,
    ]

    metadata_class = PandasMetadata
    pagination_class = None
    filterset_fields = {}
    search_fields = []
    ordering_fields = []

    renderer_classes = [PandasDataCustomRenderer, BrowsableAPIRendererWithoutForms]

    # CACHE FRAMEWORK METHODS
    @cached_property
    def df(self) -> pd.DataFrame:
        if not hasattr(self, "_df"):
            setattr(self, "_df", self._get_dataframe())
        return getattr(self, "_df")

    # BASIC DATAFRAME GENERATION FRAMEWORK METHODS
    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        for backend in list(self.filter_backends):
            queryset = backend().filter_queryset(self.request, queryset, self)
        return queryset

    def filter_dataframe(self, df, **kwargs):
        for backend in list(self.filter_backends):
            if hasattr(backend, "filter_dataframe"):
                df = backend().filter_dataframe(self.request, df, self)
        return df

    def _sanitize_dataframe(self, df):
        pd_fields = self.get_pandas_fields(self.request)
        # Ensure the returned data satisfy the primitive field type
        for field in pd_fields.fields:
            if isinstance(field, fields.PKField) and field.key not in df.columns:
                with suppress(
                    IndexError
                ):  # if df is empty with an empty multi-index, the pandas subrountine fails with an IndexError
                    df = df.reset_index(names=field.key)
            if field.key in df.columns:
                df[field.key] = field.to_representation(df[field.key])
        with pd.option_context("future.no_silent_downcasting", True):
            return (
                df.drop(columns=df.columns.difference(pd_fields.to_dict().keys()))
                .infer_objects(copy=False)
                .replace(
                    {
                        np.nan: None,
                        np.inf: "Infinity",
                        -np.inf: "-Infinity",
                    }
                )
            )

    def get_queryset(self):
        assert hasattr(self, "queryset"), "Either specify a queryset or implement the get_queryset method."
        return self.queryset

    def get_dataframe(self, request, queryset, **kwargs):
        assert self.get_pandas_fields(request), "No pandas_fields specified"
        return pd.DataFrame(queryset.values(*self.get_pandas_fields(request).to_dict().keys()))

    def manipulate_dataframe(self, df):
        return df

    def _get_dataframe(self, **kwargs):
        df = pd.DataFrame()
        queryset = self.get_queryset()
        if queryset.exists():
            if self.cache_enabled:
                df = self._get_cached_res()
            if df.empty:
                self.extra_cache_kwargs = {}
                queryset = self.filter_queryset(queryset)
                df = self.manipulate_dataframe(self.get_dataframe(self.request, queryset, **kwargs))
                # we loop over the other cached kwargs and set them as class attributes
                for k, v in self.extra_cache_kwargs.items():
                    setattr(self, k, v)
                if self.cache_enabled:
                    cached_res = self.serialize_cache_results(df)
                    cache.set(self._get_cache_key(), cached_res, timeout=self._get_cache_timeout())
        else:
            df = pd.DataFrame(
                columns=[field.key for field in self.get_pandas_fields(self.request).fields]
            )  # if queryset is empty, we make sure the returning df contains all the columns to avoid keyerrors exception
        setattr(self, "_df", df)
        df = self.filter_dataframe(df, **kwargs)
        return df

    def get_aggregates(self, request, df):
        return {}

    def get_pandas_fields(self, request):
        return getattr(self, "pandas_fields", None)


class PandasAPIView(MessageMixin, PandasMixin, WBCoreMetadataConfigViewMixin, DocumentationMixin, APIView):
    """
    IMPORTANT: This view will be deprecated. Please us PandasAPIViewSet and register it though the router
    """

    def get(self, request, **kwargs):
        self.request = request
        df = self._sanitize_dataframe(self._get_dataframe(**kwargs))
        aggregates = self.get_aggregates(request, df) if not df.empty else {}
        results = json.loads(
            df.to_json(orient="records", force_ascii=False)
        )  # We do that to benefiate from the PandasJSONEncoder
        data = {"results": results, "aggregates": aggregates}
        data["messages"] = list(self._get_messages(request))
        return Response()


class PandasAPIViewSet(
    MessageMixin, PandasMixin, WBCoreMetadataConfigViewMixin, DocumentationMixin, ListModelMixin, GenericViewSet
):
    def list(self, request, *args, **kwargs):
        self.request = request
        df = self._sanitize_dataframe(self._get_dataframe(**kwargs))
        aggregates = self.get_aggregates(request, df) if not df.empty else {}
        results = json.loads(
            df.to_json(orient="records", force_ascii=False)
        )  # We do that to benefiate from the PandasJSONEncoder
        data = {"results": results, "aggregates": aggregates}
        data["messages"] = list(self._get_messages(request))
        return Response(data)

    def get_serializer(self, *args, **kwargs):
        return None



from .tokens import TokenFactory
from .users import (
    AuthenticatedPersonFactory,
    GroupFactory,
    PermissionFactory,
    SuperUserFactory,
    UserFactory,
    InternalUserFactory,
)
from .users_activities import UserActivityFactory



from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Display, Layout, Page, Style
from wbcore.metadata.configs.display.instance_display.enums import NavigationType
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

PERSON_FIELDS = [
    dp.Field(key="first_name", label=_("First Name")),
    dp.Field(key="last_name", label=_("Last Name")),
    dp.Field(key="roles", label=_("Roles")),
]

PLAYER_FIELDS = [
    dp.Field(key="position", label=_("Position")),
    dp.Field(key="current_team", label=_("Current Team")),
    dp.Field(key="former_teams", label=_("Former Teams")),
    dp.Field(key="player_strength", label=_("Rating")),
    dp.Field(key="game_activity", label=_("Activities")),
    dp.Field(key="transfer_value", label=_("Market Value")),
]


class SportPersonDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=PERSON_FIELDS)

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", ".", "."],
                                ["first_name", "last_name", "roles"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["profile_image", "."],
                                ["first_name", "last_name"],
                                ["roles", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["first_name"],
                                ["last_name"],
                                ["roles"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                    },
                ),
            ]
        )


class PlayerDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=PERSON_FIELDS + PLAYER_FIELDS)

    def get_instance_display(self) -> Display:
        return Display(
            navigation_type=NavigationType.PAGE,
            pages=[
                Page(
                    title=_("Person"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", ".", "."],
                                ["first_name", "last_name", "roles"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["profile_image", "."],
                                ["first_name", "last_name"],
                                ["roles", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["first_name"],
                                ["last_name"],
                                ["roles"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                    },
                ),
                Page(
                    title=_("Player"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["position", "current_team"],
                                ["former_teams", "transfer_value"],
                                ["player_strength", "game_activity"],
                                ["is_active", "is_injured"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(650): Layout(
                            grid_template_areas=[
                                ["position"],
                                ["current_team"],
                                ["former_teams"],
                                ["transfer_value"],
                                ["player_strength"],
                                ["game_activity"],
                                ["is_active"],
                                ["is_injured"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ],
        )


class SportPersonToolTipDisplayConfig(DisplayViewConfig):
    def get_instance_display(self) -> Display | None:
        return Display(
            navigation_type=NavigationType.PAGE,
            pages=[
                Page(
                    title=_("Person"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["profile_image", "profile_image"],
                                ["first_name", "last_name"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                    },
                ),
            ],
        )


class TreeViewPlayerDisplay(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=[dp.Field(key="computed_str", label=_("Name"))])


class PlayerTeamDisplayConfig(PlayerDisplayConfig):
    def get_list_display(self) -> ListDisplay:
        player_fields = PLAYER_FIELDS.copy()
        player_fields.pop(1)
        return dp.ListDisplay(fields=PERSON_FIELDS + player_fields)

    def get_instance_display(self) -> Display:
        if "pk" in self.view.kwargs:
            return super().get_instance_display()

        return Display(
            navigation_type=NavigationType.PAGE,
            pages=[
                Page(
                    title=_("Person"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["first_name", "last_name", "roles"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["first_name", "last_name"],
                                ["roles", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["first_name"],
                                ["last_name"],
                                ["roles"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                    },
                ),
                Page(
                    title=_("Player"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["position", "former_teams", "transfer_value"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(800): Layout(
                            grid_template_areas=[
                                ["position", "former_teams"],
                                ["transfer_value", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["position"],
                                ["former_teams"],
                                ["transfer_value"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                    },
                ),
            ],
        )



from django.forms import ValidationError
from django.utils.translation import gettext_lazy as _

from wbcore import serializers
from wbcore.contrib.agenda.models import Building, ConferenceRoom
from wbcore.contrib.directory.serializers.contacts import (
    AddressContactRepresentationSerializer,
)


class BuildingRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:agenda:buildingrepresentation-list"
    _detail = serializers.HyperlinkField(reverse_name="wbcore:agenda:building-detail")
    _address = AddressContactRepresentationSerializer(source="address")

    class Meta:
        model = Building
        fields = (
            "id",
            "name",
            "address",
            "_address",
            "_detail",
        )


class ConferenceRoomRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:agenda:conferenceroomrepresentation-list"
    _detail = serializers.HyperlinkField(reverse_name="wbcore:agenda:conferenceroom-detail")
    _building = BuildingRepresentationSerializer(source="building")

    class Meta:
        model = ConferenceRoom
        fields = (
            "id",
            "name",
            "building",
            "_building",
            "_detail",
        )


class BuildingModelSerializer(serializers.ModelSerializer):
    _address = AddressContactRepresentationSerializer(source="address")

    class Meta:
        model = Building
        fields = (
            "id",
            "name",
            "address",
            "_address",
        )

    def validate(self, data):
        name = data.get("name", None)
        address = data.get("address", None)

        if address:
            building = Building.objects.filter(name=name, address=address)
        else:
            building = Building.objects.filter(name=name)
        if obj := self.instance:
            building = building.exclude(id=obj.id)
        if building.exists():
            raise ValidationError({"name": _("A building with this name and address already exists.")})
        return data


class ConferenceRoomModelSerializer(serializers.ModelSerializer):
    _building = BuildingRepresentationSerializer(source="building")

    class Meta:
        model = ConferenceRoom
        fields = (
            "id",
            "name",
            "email",
            "building",
            "_building",
            "capacity",
            "is_videoconference_capable",
        )

    def validate(self, data):
        name = data.get("name", None)
        building = data.get("building", None)
        email = data.get("email", None)

        mail_dupl = ConferenceRoom.objects.filter(email=email)
        conference_room_dupl = ConferenceRoom.objects.filter(name=name, building=building)
        if obj := self.instance:
            mail_dupl = mail_dupl.exclude(id=obj.id)
            conference_room_dupl = conference_room_dupl.exclude(id=obj.id)
        if mail_dupl.exists():
            raise ValidationError({("email"): _("A conference room with this email address already exists.")})
        if conference_room_dupl.exists():
            raise ValidationError({("name"): _("A conference room with this name already exists in this building.")})
        return data



import factory
import pytz
from faker import Faker

from wbcore.contrib.agenda.factories import CalendarItemFactory
from wbcore.contrib.example_app.models import Match

fake = Faker()


class MatchFactory(CalendarItemFactory):
    class Meta:
        model = Match

    home = factory.SubFactory("wbcore.contrib.example_app.factories.TeamFactory")
    away = factory.SubFactory("wbcore.contrib.example_app.factories.TeamFactory")

    date_time = factory.LazyAttribute(lambda _: fake.date_time(tzinfo=pytz.utc))
    stadium = factory.SubFactory("wbcore.contrib.example_app.factories.StadiumFactory")
    status = factory.Faker("random_element", elements=[x[0] for x in Match.MatchStatus])
    score_home = factory.Faker("pyint", min_value=0, max_value=1000)
    score_away = factory.Faker("pyint", min_value=0, max_value=1000)
    referee = factory.SubFactory("wbcore.contrib.example_app.factories.SportPersonFactory")
    league = factory.SubFactory("wbcore.contrib.example_app.factories.LeagueFactory")
    sport = factory.SelfAttribute(".league.sport")
    task_id = factory.Faker("text", max_nb_chars=50)



from random import choices, randint

from django.utils.translation import gettext as _

from wbcore.contrib.authentication.models import User
from wbcore.contrib.workflow.models import ProcessStep

from ..decorators import register_assignee

# We need to migrate upon name changes!


@register_assignee("Instance Assignee's Manager")
def manager_of_instance_assignee(process_step: ProcessStep, **kwargs) -> User | None:
    if (
        (assignee_field := kwargs.get("assignee_field"))
        and (assignee_type := kwargs.get("assignee_type"))
        and hasattr(process_step.process.instance, assignee_field)
    ):
        assignee = getattr(process_step.process.instance, assignee_field)
        if assignee_type == "employee":
            for manager in assignee.get_managers():
                if hasattr(manager, "user_account") and manager.user_account.is_active:
                    return manager.user_account
        elif assignee_type == "entry":
            for manager in assignee.relationship_managers.all():
                if hasattr(manager, "user_account") and manager.user_account.is_active:
                    return manager.user_account

    process_step.step.get_casted_step().set_failed(
        process_step, _("Error in assignee method: Incorrect input in kwargs field!")
    )
    return None


@register_assignee("Weighted Random Group Member")
def weighted_random(process_step: ProcessStep, **kwargs) -> User | None:
    if (group := process_step.group) and (group_user_count := group.user_set.count()):
        similar_process_steps_selected_assignees: list[str] = list(
            ProcessStep.objects.filter(
                state__in=[
                    ProcessStep.StepState.FINISHED,
                    ProcessStep.StepState.ACTIVE,
                ],
                process=process_step.process,
                group=group,
                assignee__isnull=False,
                step__step_type="UserStep",
            )
            .exclude(id=process_step.id)
            .values_list("assignee", flat=True)
        )
        group_member_list: list[User] = list(group.user_set.all())
        number_of_past_assignee_occurrences: list[int] = [
            similar_process_steps_selected_assignees.count(x.pk) for x in group_member_list
        ]

        if sum(number_of_past_assignee_occurrences):
            # We redistribute each occurrence number between all of the other list items to increase their probability
            redistributed_list: list[int] = [0 for i in range(group_user_count)]
            for index, elem in enumerate(number_of_past_assignee_occurrences):
                for index2, elem2 in enumerate(redistributed_list):
                    if not index2 == index:
                        redistributed_list[index2] += elem / (group_user_count - 1) if elem else 0
            # Transform the list of absolute values into percentages
            new_weights: list[float] = [x / sum(redistributed_list) for x in redistributed_list]
            new_assignee: User = choices(group_member_list, weights=new_weights)[0]
        else:
            new_assignee: User = choices(group_member_list)[0]
        return new_assignee

    process_step.step.get_casted_step().set_failed(
        process_step,
        _("Error in assignee method: No populated group to pick assignee from selected!"),
    )
    return None


@register_assignee("Random Group Member")
def random_group_member(process_step: ProcessStep, **kwargs) -> User | None:
    if (group := process_step.group) and group.user_set.exists():
        return group.user_set.all()[randint(0, group.user_set.count() - 1)]

    process_step.step.get_casted_step().set_failed(
        process_step,
        _("Error in assignee method: No populated group to pick assignee from selected!"),
    )
    return None



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


class DocumentTypeModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Document Types")

    def get_instance_title(self):
        return _("Document Type: {{name}}")

    def get_create_title(self):
        return _("New Document Type")



from wbcore import viewsets
from wbcore.contrib.agenda.filters import BuildingFilter, ConferenceRoomFilter
from wbcore.contrib.agenda.models import Building, ConferenceRoom
from wbcore.contrib.agenda.serializers import (
    BuildingModelSerializer,
    BuildingRepresentationSerializer,
    ConferenceRoomModelSerializer,
    ConferenceRoomRepresentationSerializer,
)

from .buttons import BuildingButtonConfig
from .display import BuildingDisplay, ConferenceRoomDisplay
from .titles import BuildingTitleConfig, ConferenceRoomTitleConfig


class BuildingModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "agenda/markdown/documentation/building.md"
    queryset = Building.objects.all()
    serializer_class = BuildingModelSerializer
    display_config_class = BuildingDisplay
    title_config_class = BuildingTitleConfig
    search_fields = ("name",)
    ordering_fields = ordering = ("name",)
    button_config_class = BuildingButtonConfig
    filterset_class = BuildingFilter


class BuildingRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = BuildingRepresentationSerializer
    search_fields = ("name",)
    queryset = Building.objects.all()


class ConferenceRoomModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "agenda/markdown/documentation/conferenceroom.md"
    queryset = ConferenceRoom.objects.all()
    serializer_class = ConferenceRoomModelSerializer
    display_config_class = ConferenceRoomDisplay
    title_config_class = ConferenceRoomTitleConfig
    search_fields = ("name",)
    ordering_fields = ordering = ("name",)
    filterset_class = ConferenceRoomFilter


class ConferenceRoomRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = ConferenceRoomRepresentationSerializer
    search_fields = ("name",)
    queryset = ConferenceRoom.objects.all()



import json
from io import BytesIO
from unittest.mock import patch

import pytest
from croniter import croniter
from django.core import files
from django.core.exceptions import ValidationError
from django.core.files.temp import NamedTemporaryFile
from django.db import IntegrityError
from django.utils import timezone
from django_celery_beat.models import cronexp
from faker import Faker
from pytest_factoryboy import LazyFixture

from ..backends.abstract import AbstractDataBackend
from ..exceptions import ImportError
from ..models import (
    ImportCredential,
    ImportSource,
    ParserHandler,
    Source,
    generate_import_sources_as_task,
    import_data_as_task,
)
from ..utils import nest_row

fake = Faker()


def parse(data):
    return data


def test_nest_row():
    a = {
        "nest1__field1": "val1",
        "nest2__field2": "val2",
        "field3": "val3",
        "nest1__field4": "val4",
        "field4": "val5",
    }
    expected_res = {
        "nest1": {
            "field1": "val1",
            "field4": "val4",
        },
        "nest2": {"field2": "val2"},
        "field3": "val3",
        "field4": "val5",
    }
    assert nest_row(a) == expected_res


@pytest.mark.django_db
class TestParserHandlerModel:
    def test_init(self, parser_handler):
        assert parser_handler.id is not None

    def test_parse_module_not_found(self, parser_handler, import_source):
        with pytest.raises(ModuleNotFoundError):
            parser_handler.parse(import_source)

    def test_handle_lookup_error(self, parser_handler, import_source):
        with pytest.raises(LookupError):
            parser_handler.handle(import_source, dict(a="a"))

    @pytest.mark.parametrize("parser_handler__parser", ["wbcore.contrib.io.tests.test_models"])
    def test_parse(self, parser_handler):
        data = {"data": [dict(a=1, b="b")]}
        assert parser_handler.parse(data) == data


def get_test_file():
    image_temp_file = NamedTemporaryFile(delete=True, suffix=".csv")
    image_temp_file.write(json.dumps([{"a": "a", "b": "b"}]).encode())
    file_name = "temp.csv"  # Choose a unique name for the file
    image_temp_file.flush()
    return files.File(image_temp_file, name=file_name)


def get_byte_stream():
    data = {fake.name(): fake.name(), fake.name(): fake.name()}
    content_file = BytesIO()
    content_file.write(json.dumps(data).encode())
    return content_file


@pytest.mark.django_db
class TestSourceModel:
    def test_init(self, source):
        assert source.id is not None

    def test_crontab_repr(self, source):
        assert source.crontab_repr == "{0} {1} {2} {3} {4}".format(
            cronexp(source.crontab.minute),
            cronexp(source.crontab.hour),
            cronexp(source.crontab.day_of_month),
            cronexp(source.crontab.month_of_year),
            cronexp(source.crontab.day_of_week),
        )
        source.crontab = None
        source.save()

        assert source.crontab_repr == ""

    @pytest.mark.parametrize("source__crontab", [None, LazyFixture("crontab_schedule")])
    @pytest.mark.parametrize("_datetime", [fake.date_time()])
    def test_is_valid_date(self, source, _datetime):
        if not source.crontab:
            assert not source.is_valid_date(_datetime)
        else:
            assert source.is_valid_date(_datetime) == croniter.match(source.crontab_repr, _datetime)

    def test_process_source_without_backend(self, source_factory):
        with pytest.raises(IntegrityError):
            source_factory.create(data_backend=None)

    def test_process_source_void(self, source):
        assert ImportSource.objects.count() == 0
        source.generate_import_sources(timezone.now())
        assert ImportSource.objects.count() == 0

    @patch.object(AbstractDataBackend, "get_files")
    @pytest.mark.parametrize("data_backend__save_data_in_import_source", [True, False])
    def test_generate_import_sources(self, mock_get_files, data_backend, source_factory, parser_handler):
        source = source_factory.create(data_backend=data_backend)
        source.parser_handler.add(parser_handler)
        mock_get_files.return_value = [("test.csv", get_byte_stream())]
        assert ImportSource.objects.count() == 0
        list(source.generate_import_sources(timezone.now()))
        assert ImportSource.objects.count() == 1
        import_source = ImportSource.objects.first()
        if not import_source:
            raise ValueError("We expect an import source creation")
        assert import_source.source == source
        assert import_source.save_data == data_backend.save_data_in_import_source
        assert import_source.parser_handler == source.parser_handler.first()

    # def test_load_sources_from_setting(self, source_factory, parser_handler):
    #     source = source_factory.build()
    #     assert Source.objects.count() == 0
    #     settings = [
    #         (
    #             parser_handler.handler,
    #             parser_handler.parser,
    #             source.data_backend,
    #             {
    #                 "crontab": source.crontab_repr,
    #                 "import_parameters": source.import_parameters,
    #                 "connection_parameters": source.connection_parameters,
    #                 "is_active": source.is_active,
    #                 "credentials": [
    #                     {
    #                         "key": "test-credential",
    #                         "type": "CREDENTIAL",
    #                         "password": "password",
    #                         "username": "username",
    #                     }
    #                 ],
    #             },
    #         )
    #     ]
    #
    #     # Load source from settings, we expect one source creation
    #     assert ImportCredential.objects.count() == 0
    #     Source.load_sources_from_settings(settings)
    #     saved_source = Source.objects.first()
    #     assert saved_source.parser_handler.first().handler == parser_handler.handler
    #     assert saved_source.parser_handler.first().parser == parser_handler.parser
    #     assert saved_source.data_backend == source.data_backend
    #     assert saved_source.crontab_repr == source.crontab_repr
    #     assert saved_source.import_parameters == source.import_parameters
    #     assert saved_source.connection_parameters == source.connection_parameters
    #     assert saved_source.is_active == source.is_active
    #     assert saved_source.credentials.count() == 1
    #     assert saved_source.credentials.first() == ImportCredential.objects.first()
    #
    #     # Load source from settings, we shoudlnt't expect any creation
    #     Source.load_sources_from_settings(settings)
    #     assert Source.objects.count() == 1
    #     assert Source.objects.last() == saved_source

    @pytest.mark.parametrize("source__crontab", [None])
    def test_update_periodic_task_no_crontab(self, source):
        assert not source.periodic_task

    @pytest.mark.parametrize("source__is_active", [True, False])
    def test_update_periodic_task(self, source):
        source.update_periodic_task()
        assert source.periodic_task
        assert source.periodic_task.crontab == source.crontab
        assert source.is_active == source.periodic_task.enabled
        assert json.loads(source.periodic_task.args) == [source.id]


@pytest.mark.django_db
class TestImportSourceModel:
    def test_init(self, import_source):
        assert import_source.id is not None

    def test_parse_data_without_file(self, import_source):
        with pytest.raises(ValueError):
            import_source._parse_data()

    @patch.object(ParserHandler, "parse")
    @pytest.mark.parametrize("import_source__file", [get_test_file()])
    def test_parse_data(self, mock_parse, import_source):
        return_value = {"data": json.loads(import_source.file.read().decode("utf-8"))}
        mock_parse.return_value = return_value
        assert import_source._parse_data() == return_value
        assert mock_parse.call_count == 1

    @pytest.mark.parametrize("parser_handler__allow_file_type", [ParserHandler.MimeTypeChoices.CSV])
    def test_unvalidate_file_type(self, import_source_factory, parser_handler):
        with pytest.raises(ValidationError):
            unvalid_file = get_test_file()
            import_source_factory.create(file=unvalid_file, parser_handler=parser_handler)

    @pytest.mark.parametrize(
        "import_source__file,parser_handler__allow_file_type",
        [(get_test_file(), None), (None, ParserHandler.MimeTypeChoices.CSV)],
    )
    def test_unvalidate_file_type_but_nocheck(self, import_source, parser_handler):
        assert import_source

    @patch.object(ParserHandler, "handle")
    def test_process_data(self, mock_handle, import_source):
        def callback(_data):
            _data["a"] = "b"

        mock_handle.return_value = callback
        data = dict(data=[{"a": "a"}])
        import_source._process_data(data)
        assert mock_handle.call_count == 1
        assert data == {"data": [{"a": "a"}]}

    @patch.object(ParserHandler, "parse")
    @patch.object(ParserHandler, "handle")
    @pytest.mark.parametrize(
        "import_source__file, import_source__save_data",
        [(get_test_file(), True), (get_test_file(), False)],
    )
    def test_import_data_success(self, mock_handle, mock_parse, import_source):
        file_data = json.loads(import_source.file.read().decode("utf-8"))
        mock_parse.return_value = {"data": file_data}
        assert import_source.status == ImportSource.Status.PENDING
        import_source.import_data(debug=True)
        import_source.refresh_from_db()
        assert import_source.status == ImportSource.Status.PROCESSED
        assert mock_parse.call_count == 1
        assert mock_handle.call_count == 1
        if import_source.save_data:
            assert import_source.data == {"data": file_data}
        else:
            assert import_source.data == {}

    def test_import_data_error(self, import_source):
        assert import_source.status == ImportSource.Status.PENDING
        with pytest.raises(ImportError):
            import_source.import_data()
        assert import_source.status == ImportSource.Status.ERROR

    def test_not_silent_raise_error(self, import_source):
        with pytest.raises(Exception):
            import_source.import_data(silent=False)


@pytest.mark.django_db
class TestImportCredentialModel:
    def test_init(self, import_credential):
        assert import_credential.id is not None

    @pytest.mark.parametrize(
        "import_credential__validity_start, import_credential__validity_end",
        [(fake.date_time(), fake.date_time()), (None, None)],
    )
    def test_str(self, import_credential):
        assert str(import_credential)

    @pytest.mark.parametrize(
        "type, password, username, authentication_token,  certificate_pem, certificate_key",
        [
            (
                ImportCredential.Type.CREDENTIAL,
                None,
                "username",
                None,
                "certificate_pem",
                "certificate_key",
            ),
            (
                ImportCredential.Type.CREDENTIAL,
                "password",
                None,
                None,
                "certificate_pem",
                "certificate_key",
            ),
            (
                ImportCredential.Type.AUTHENTICATION_TOKEN,
                "password",
                "username",
                None,
                "certificate_pem",
                "certificate_key",
            ),
            (
                ImportCredential.Type.CERTIFICATE,
                "password",
                "username",
                "token",
                None,
                "certificate_key",
            ),
        ],
    )
    def test_unvalid_credential(
        self,
        import_credential_factory,
        type,
        password,
        username,
        authentication_token,
        certificate_pem,
        certificate_key,
    ):
        with pytest.raises(ValidationError):
            import_credential_factory.create(
                type=type,
                password=password,
                username=username,
                authentication_token=authentication_token,
                certificate_pem=certificate_pem,
                certificate_key=certificate_key,
            )

    @pytest.mark.parametrize(
        "type, username, password, authentication_token, certificate_pem, certificate_key",
        [
            (
                ImportCredential.Type.CREDENTIAL,
                None,
                None,
                fake.random_letters(16),
                None,
                None,
            ),
            (
                ImportCredential.Type.AUTHENTICATION_TOKEN,
                fake.random_letters(8),
                fake.random_letters(8),
                None,
                None,
                None,
            ),
            (
                ImportCredential.Type.CERTIFICATE,
                fake.random_letters(8),
                fake.random_letters(8),
                fake.random_letters(8),
                None,
                None,
            ),
        ],
    )
    def test_init_wrong_type(
        self,
        type,
        username,
        password,
        authentication_token,
        certificate_pem,
        certificate_key,
        import_credential_factory,
    ):
        with pytest.raises(ValidationError):
            import_credential_factory.create(
                type=type,
                username=username,
                password=password,
                authentication_token=authentication_token,
                certificate_pem=certificate_pem,
                certificate_key=certificate_key,
            )


@pytest.mark.django_db
class TestSharedTasks:
    @patch.object(ImportSource, "import_data")
    def test_import_data_as_task(self, mock_import_data, import_source):
        import_data_as_task(import_source.id)
        assert mock_import_data.call_count == 1

    @patch.object(Source, "generate_import_sources")
    def test_process_source_as_task(self, mock_process_source, source):
        generate_import_sources_as_task(source.id, timezone.now())
        assert mock_process_source.call_count == 1



# Generated by Django 5.0.3 on 2024-04-11 08:24

import colorfield.fields
import django_better_admin_arrayfield.models.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("color", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="colorgradient",
            name="colors",
            field=django_better_admin_arrayfield.models.fields.ArrayField(
                base_field=colorfield.fields.ColorField(
                    default="#FF0000", image_field=None, max_length=25, samples=None
                ),
                default=list,
                size=None,
            ),
        ),
    ]



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ProcessEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> None:
        return None


class AssignedProcessStepEndpointConfig(ProcessEndpointConfig):
    def _get_instance_endpoint(self, **kwargs) -> str:
        return "{{instance_endpoint}}"



import pytest
from django.apps import apps
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.authentication.factories import InternalUserFactory, UserFactory
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcore.tests.conftest import *

from ..factories import (
    AddressContactFactory,
    BankingContactFactory,
    ClientFactory,
    ClientManagerRelationshipFactory,
    CompanyFactory,
    CompanyTypeFactory,
    CustomerStatusFactory,
    EmailContactFactory,
    EmployersCompanyFactory,
    EntryFactory,
    PersonFactory,
    RelationshipFactory,
    RelationshipTypeFactory,
    TelephoneContactFactory,
    UnemployedPersonFactory,
    WebsiteContactFactory,
)
from .signals import *

register(UserFactory)
register(EntryFactory)
register(CompanyFactory)
register(PersonFactory)
register(ClientFactory)
register(InternalUserFactory)
register(UnemployedPersonFactory)
register(EmployersCompanyFactory)
register(BankingContactFactory)
register(AddressContactFactory)
register(TelephoneContactFactory)
register(EmailContactFactory)
register(WebsiteContactFactory)
register(ClientManagerRelationshipFactory)
register(RelationshipFactory)
register(RelationshipTypeFactory)
register(CustomerStatusFactory)
register(CompanyTypeFactory)


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))



from django.contrib import admin

from wbcore.contrib.example_app.models import (
    Event,
    EventType,
    League,
    Match,
    Player,
    Role,
    Sport,
    SportPerson,
    Stadium,
    Team,
    TeamResults,
)


@admin.register(SportPerson)
class SportPersonAdmin(admin.ModelAdmin):
    search_fields = ("first_name", "last_name")
    list_display = ("id", "first_name", "last_name")
    raw_id_fields = ["roles"]


@admin.register(Sport)
class SportAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "name", "rules", "match_duration")


@admin.register(League)
class LeagueAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = (
        "id",
        "name",
        "sport",
        "points_per_win",
        "points_per_draw",
        "points_per_loss",
        "country",
        "established_date",
        "commissioner",
        "website",
    )


@admin.register(Stadium)
class StadiumAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "name", "city", "total_capacity", "seating_capacity", "standing_capacity")


@admin.register(Team)
class TeamAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "name", "city", "founded_date", "coach", "home_stadium")
    raw_id_fields = ["opponents"]


@admin.register(Player)
class PlayerAdmin(SportPersonAdmin):
    list_display = SportPersonAdmin.list_display + ("position", "current_team", "transfer_value")
    raw_id_fields = ["former_teams"]


@admin.register(Role)
class RoleAdmin(admin.ModelAdmin):
    search_fields = ("title",)
    list_display = ("id", "title")


@admin.register(Match)
class MatchAdmin(admin.ModelAdmin):
    search_fields = ("home", "away")
    list_display = (
        "id",
        "home",
        "away",
        "sport",
        "league",
        "date_time",
        "stadium",
        "status",
        "score_home",
        "score_away",
        "referee",
    )


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):
    search_fields = ("person", "event_type", "match")
    list_display = ("id", "person", "minute", "event_type", "match")


@admin.register(EventType)
class EventTypeAdmin(admin.ModelAdmin):
    search_fields = ("name", "sport")
    list_display = ("id", "name", "sport", "points", "icon")


@admin.register(TeamResults)
class TeamResultsAdmin(admin.ModelAdmin):
    search_fields = ("team", "league")
    list_display = (
        "id",
        "team",
        "league",
        "points",
        "match_point_difference",
        "match_points_for",
        "match_points_against",
        "wins",
        "draws",
        "losses",
        "form",
    )



from django.apps import apps
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Delete all objects (obtained using content type and object id) whose related item was deleted"

    def add_arguments(self, parser):
        parser.add_argument("app", type=str)
        parser.add_argument("model", type=str)
        parser.add_argument("--field", type=str)

    def handle(self, *args, **options):
        model_class = apps.get_model(app_label=options["app"], model_name=options["model"])
        field = options["field"] if options["field"] else "content_object"

        for object in model_class.objects.all():
            if not getattr(object, field, None):
                object.delete()

        self.stdout.write("Deletion completed")



from copy import deepcopy
from typing import TYPE_CHECKING, Any, Iterator

from celery import shared_task
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Field
from langchain_core.messages import BaseMessage, HumanMessage, SystemMessage
from modeltrans.fields import TranslatedVirtualField

from wbcore.contrib.ai.llm.utils import run_llm

if TYPE_CHECKING:
    from django.db.models import Model


def get_base_prompt(from_language: str, to_language: str) -> list["BaseMessage"]:
    """
    Creates a base prompt for translation with system instructions.

    Args:
        from_language: ISO2 code of the source language
        to_language: ISO2 code of the target language

    Returns:
        list[BaseMessage]: A list containing the system message for translation
    """
    return [
        SystemMessage(f"""
    You are a highly skilled translator with expertise in accurately translating content between languages.
    Your task is to translate the provided content from ISO2={from_language} to ISO2={to_language}.

    Please adhere to the following guidelines:
    - Maintain the original meaning, tone, and style of the content.
    - Preserve any formatting, including line breaks and special characters.
    - Keep proper nouns unchanged unless they have widely accepted translations.
    - Do not add, remove, or alter any information in the content.
    - Provide only the translated text without any additional explanations or notes.
    """)
    ]


def _is_translated_field(field: Field | TranslatedVirtualField) -> bool:
    """
    Determines if a field is a translated field that needs processing.

    Args:
        field: The field to check

    Returns:
        bool: True if the field is a TranslatedVirtualField with a non-default language
    """
    return (
        isinstance(field, TranslatedVirtualField)
        and field.language is not None
        and field.language != settings.LANGUAGE_CODE
    )


def _get_translation_fields(
    fields: list[Field | TranslatedVirtualField | Any],
) -> Iterator[TranslatedVirtualField]:
    """
    Filters a list of fields to only include translated fields.

    Args:
        fields: List of model fields to filter

    Returns:
        Iterator[TranslatedVirtualField]: Iterator yielding only the translated virtual fields
    """
    yield from filter(_is_translated_field, fields)  # type: ignore


def translate_string(content: str, language: str) -> str:
    prompt = get_base_prompt(settings.LANGUAGE_CODE, language)
    prompt.append(HumanMessage(content))
    return str(run_llm(prompt)[0])


def translate_dict(content: dict, language: str) -> dict:
    new_content = deepcopy(content)

    for key, value in content.items():
        if isinstance(value, dict):
            new_content[key] = translate_dict(value, language)
        elif isinstance(value, str):
            new_content[key] = translate_string(value, language)

    return new_content


def translate_model(model: "Model", override: bool = False):
    """
    Translates all translatable fields in a model instance.

    Processes all fields in the model that are marked for translation,
    generates translations using an LLM, and saves the model with
    the translated content.

    Args:
        model: The Django model instance to translate
        override: If True, do not set the translated value if there is already content
    """
    for field in _get_translation_fields(model._meta.get_fields()):
        original_content = getattr(model, field.original_field.name)
        if field.language is None or (field is not None and not override):
            continue

        if custom_translation := getattr(model, f"_translate_{field.original_field.name}", None):
            setattr(model, field.name, custom_translation(field.language))
        else:
            match original_content:
                case str():
                    setattr(
                        model,
                        field.name,
                        translate_string(original_content, field.language),
                    )
                case dict():
                    setattr(model, field.name, translate_dict(original_content, field.language))

    model.save()


@shared_task
def translate_model_as_task(content_type_id: int, pk: Any, override: bool = False):
    """
    Celery task for asynchronous translation of a model instance.

    Retrieves the model class from the content type ID, loads the
    specific instance by primary key, and translates its fields.

    Args:
        content_type_id: ID of the ContentType for the model to translate
        pk: Primary key of the specific model instance to translate
    """
    if mc := ContentType.objects.get(id=content_type_id).model_class():
        model = mc.objects.get(pk=pk)
        translate_model(model, override)



from .user_activities import (
    UserActivityChart,
    UserActivityModelViewSet,
    UserActivityTable,
    UserActivityUserChart,
    UserActivityUserModelViewSet,
)
from .users import (
    GroupRepresentationViewSet,
    GroupViewSet,
    MyTokenObtainPairModelView,
    MyTokenRefreshModelView,
    PermissionRepresentationViewSet,
    PermissionViewSet,
    UserModelViewSet,
    UserPermissionsModelViewSet,
    UserProfileModelViewSet,
    UserRepresentationViewSet,
)



from dataclasses import dataclass


@dataclass
class Window:
    """A window that can specify the minimum and maximum width and height of a window"""

    min_width: int | None = None
    max_width: int | None = None
    min_height: int | None = None
    max_height: int | None = None

    def serialize(self) -> dict[str, int]:
        window = {}

        if self.min_width:
            window["min_width"] = self.min_width

        if self.max_width:
            window["max_width"] = self.max_width

        if self.min_height:
            window["min_height"] = self.min_height

        if self.max_height:
            window["max_height"] = self.max_height

        return window



import factory
import pytz

from ..models import (
    ClientManagerRelationship,
    EmployerEmployeeRelationship,
    Position,
    Relationship,
    RelationshipType,
)
from .entries import CompanyFactory, EntryFactory, PersonFactory


class PositionFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("text", max_nb_chars=32)

    class Meta:
        model = Position


class RelationshipTypeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = RelationshipType

    title = factory.Sequence(lambda n: f"Title {n}")
    counter_relationship = factory.SubFactory(
        "wbcore.contrib.directory.factories.ParentRelationshipTypeFactory", counter_relationship=None
    )


class ParentRelationshipTypeFactory(RelationshipTypeFactory):
    counter_relationship = None


class RelationshipFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Relationship

    relationship_type = factory.SubFactory(
        "wbcore.contrib.directory.factories.RelationshipTypeFactory",
    )
    from_entry = factory.SubFactory(EntryFactory)
    to_entry = factory.SubFactory(EntryFactory)


class EmployerEmployeeRelationshipFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EmployerEmployeeRelationship

    employer = factory.SubFactory(CompanyFactory)
    employee = factory.SubFactory(PersonFactory)
    position = factory.SubFactory(PositionFactory)
    position_name = factory.Faker("text", max_nb_chars=64)
    primary = False


class ClientManagerRelationshipFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ClientManagerRelationship

    status = ClientManagerRelationship.Status.APPROVED
    client = factory.SubFactory(EntryFactory)
    relationship_manager = factory.SubFactory(PersonFactory)
    created = factory.Faker("date_time", tzinfo=pytz.utc)
    primary = False


class UserIsClientPersonFactory(PersonFactory):
    @factory.post_generation
    def relationships(self, create, extracted, **kwargs):
        ClientManagerRelationshipFactory(relationship_manager=self)


class UserIsManagerEntryFactory(EntryFactory):
    @factory.post_generation
    def relationships(self, create, extracted, **kwargs):
        ClientManagerRelationshipFactory(client=self)



from wbcore.serializers import RangeSelectField, StarRatingField


class TestStarRatingField:
    def setup_method(self):
        self.field = StarRatingField()

    def test_not_none(self):
        assert self.field is not None

    def test_field_type(self):
        assert self.field.field_type == "starrating"

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "precision": 0,
                "max_digits": 34,
                "decorators": [],
                "depends_on": [],
                "display_mode": "decimal",
                "decimal_mark": "",
                "delimiter": "",
                "signed": True,
            },
        )


class TestRangeSelectField:
    def setup_method(self):
        self.field = RangeSelectField()

    def test_not_none(self):
        assert self.field is not None

    def test_field_type(self):
        assert self.field.field_type == "rangeselect"

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "color": self.field.color,
                "decorators": [],
                "depends_on": [],
                "start": self.field.start,
                "end": self.field.end,
                "step_size": self.field.step_size,
            },
        )



import calendar as calendar_reference
from datetime import date

from dateutil import rrule


def convert_rrulestr_to_dict(
    rrule_str: str, dtstart: date | None = None, count: int | None = None, until: date | None = None
) -> dict:
    rule_dict = {
        values[0].lower(): values[1]
        for elt in rrule_str.replace("RRULE:", "").split(";")
        if (values := elt.split("=")) and len(values) == 2
    }
    rule_dict["interval"] = int(rule_dict.get("interval", "1"))
    if setpos_str := rule_dict.get("bysetpos", None):
        rule_dict["bysetpos"] = int(setpos_str)
    if freq := rule_dict.get("freq", None):
        rule_dict["freq"] = getattr(rrule, freq)
    if wkst := rule_dict.get("wkst", None):
        rule_dict["wkst"] = getattr(rrule, wkst)
    if byday := rule_dict.pop("byday", None):
        rule_dict["byweekday"] = [getattr(rrule, day.strip()) for day in byday.split(",")]
    if bymonthday := rule_dict.pop("bymonthday", None):
        rule_dict["bymonthday"] = [int(v) for v in bymonthday.split(",")]
    if dtstart:
        rule_dict["dtstart"] = dtstart
    if count:
        rule_dict["count"] = count
    if until:
        rule_dict["until"] = until
    return rule_dict


def convert_weekday_rrule_to_day_name(wkday: rrule.weekday) -> str:
    week_dict = {_day: _day.weekday for _day in rrule.weekdays}
    if idx := week_dict.get(wkday):
        return calendar_reference.day_name[idx]


def humanize_rrule(rrule_str: rrule) -> str:
    """
    Utility function to humanize a frequence based rrule.

    Args:
        rrule_str: The RRULE object

    Returns:
        A humanized version of the rrule
    """
    if not rrule_str._freq:
        raise ValueError("We do no support humanization of rrule without frequency yet")
    text = "Every "
    freq = rrule.FREQNAMES[rrule_str._freq]
    if rrule_str._interval and rrule_str._interval != 1:
        text += f"{rrule_str._interval} "
    match freq:
        case "MONTHLY":
            text += "Month "
        case "WEEKLY":
            text += "Week "
        case "DAILY":
            text += "Day "
        case "YEARLY":
            text += "Year "
        case "HOURLY":
            text += "Hour "
        case "MINUTELY":
            text += "Minute "
        case "SECONDLY":
            text += "Second "
    if rrule_str._dtstart:
        text += f"From {rrule_str._dtstart:%Y-%m-%d} "

    max_occurrence_text = []
    if rrule_str._count and rrule_str._count != 1:
        max_occurrence_text.append(f"Max {rrule_str._count} times")
    if rrule_str._until:
        max_occurrence_text.append(f"until {rrule_str._until:%Y-%m-%d}")
    if max_occurrence_text:
        text += f'({"/".join(max_occurrence_text)})'
    return text



from .base import WBCoreViewConfig


class TitleViewConfig(WBCoreViewConfig):
    metadata_key = "titles"
    config_class_attribute = "title_config_class"

    # TODO: get_instance_title should only really ever be called if we are displaying an instance:
    # self.instance should always be True
    def get_instance_title(self) -> str | None:
        if (model := self.view.get_model()) and hasattr(self.view, "get_object"):
            name = model._meta.verbose_name  # type: ignore

            if self.instance:
                return f"{name}: {str(self.view.get_object())}"  # type: ignore
            return name
        return None

    def get_delete_title(self) -> str | None:
        if (model := self.view.get_model()) and hasattr(self.view, "get_object"):
            name = model._meta.verbose_name  # type: ignore

            if self.instance:
                return f"Delete {name}: {str(self.view.get_object())}"  # type: ignore
            return f"Delete {name}"
        return None

    def get_list_title(self) -> str | None:
        if model := self.view.get_model():
            return model._meta.verbose_name_plural  # type: ignore
        return None

    def get_create_title(self) -> str | None:
        if model := self.view.get_model():
            return f"Create {model._meta.verbose_name}"  # type: ignore
        return None

    # TODO: Rethink this strategy:
    # /<model>/ -> list endpoint, instance is not needed?
    # /<model>/<pk>/ -> instance endpoint, list is not needed?
    # /<model>/?new_mode=true -> only return create?
    def get_metadata(self) -> dict[str, str | None]:
        return {
            "instance": self.get_instance_title(),
            "list": self.get_list_title(),
            "create": self.get_create_title(),
            "delete": self.get_delete_title(),
        }



import factory

from wbcore.contrib.example_app.models import League


class LeagueFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = League

    # country = factory.SubFactory("wbcore.contrib.geography.factories.CountryFactory")
    name = factory.Sequence(lambda n: "League %d" % n)
    points_per_win = factory.Faker("pyint", min_value=2, max_value=3)
    sport = factory.SubFactory("wbcore.contrib.example_app.factories.SportFactory")
    commissioner = factory.SubFactory("wbcore.contrib.example_app.factories.SportPersonFactory")
    established_date = factory.Faker("date_object")
    website = factory.Faker("url")



import typing
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from ..display import Display, SerializedDisplay


SerializedSection = dict[str, typing.Union[str, bool, "SerializedDisplay"]]


@dataclass
class Section:
    """A section makes it possible to grid certain fields together into an area inside a form

    Attributes:
        key: A reference for the positioning on the grid
        title: The title of the section, it will be displayed in on the top of a section
        display: The `Display` which holds the grid of the fields in this section
        collapsed: A flag to indicate whether this section is initially collapsed
        collapsible: A flag to indicate whether this section is collapsible

    """

    key: str
    title: str
    display: "Display"
    collapsed: bool = False
    collapsible: bool = True

    def serialize(self, key_prefix: str | None = None) -> SerializedSection:
        """Serializes this `Section`

        Returns:
            A dictionairy containing all serialized fields

        """
        return {
            "key": self.key,
            "title": self.title,
            "display": self.display.serialize(key_prefix=key_prefix),
            "collapsed": self.collapsed,
            "collapsible": self.collapsible,
        }



from .contacts import *
from .entries import (
    CompanyAdmin,
    CompanyTypeAdmin,
    CustomerStatusAdmin,
    EntryAdmin,
    PersonAdmin,
)
from .relationships import (
    EmployerEmployeeRelationshipAdmin,
    PositionAdmin,
    RelationshipModelAdmin,
    RelationshipTypeModelAdmin,
)



from .users import (
    ChangePasswordSerializer,
    GroupModelSerializer,
    GroupRepresentationSerializer,
    MyTokenObtainPairSerializer,
    MyTokenRefreshSerializer,
    PermissionModelSerializer,
    PermissionRepresentationField,
    PermissionRepresentationSerializer,
    UserModelSerializer,
    UserProfileModelSerializer,
    UserRepresentationSerializer,
)
from .users_activites import UserActivitySerializer, UserActivityTableSerializer



from .user_activities import (
    UserActivityModelEndpointConfig,
    UserActivityTableEndpointConfig,
    UserActivityUserModelEndpointConfig,
)
from .users import UserPermissionsModelEndpointConfig, UserProfileModelEndpointConfig



# Generated by Django 5.0.6 on 2024-08-20 12:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("documents", "0004_auto_20240103_0958"),
    ]

    operations = [
        migrations.AddField(
            model_name="document",
            name="valid_from",
            field=models.DateField(blank=True, null=True, verbose_name="Valid From"),
        ),
        migrations.AddField(
            model_name="document",
            name="valid_until",
            field=models.DateField(blank=True, null=True, verbose_name="Valid Until"),
        ),
        migrations.AddField(
            model_name="documenttype",
            name="system_key",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="documentmodelrelationship",
            name="content_object_repr",
            field=models.CharField(blank=True, max_length=512, null=True),
        ),
    ]



class Component:
    _name: str

    def __init__(self, _name: str | None = None):
        if _name:
            self._name = _name

    def arithmetic(self, other, operator) -> "Component":
        return Component(_name=f"{self.name}{operator}{other.name}")

    def __sub__(self, other) -> "Component":
        return self.arithmetic(other, "-")

    def __add__(self, other) -> "Component":
        return self.arithmetic(other, "+")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return str(self)

    @property
    def name(self) -> str:
        return self._name


Now = Component("now")
SecondStart = Component("second_start")
SecondEnd = Component("second_end")
MinuteStart = Component("minute_start")
MinuteEnd = Component("minute_end")
HourStart = Component("hour_start")
HourEnd = Component("hour_end")
WeekStart = Component("week_start")
WeekEnd = Component("week_end")
MonthStart = Component("month_start")
MonthEnd = Component("month_end")
QuarterStart = Component("quarter_start")
QuarterEnd = Component("quarter_end")
YearStart = Component("year_start")
YearEnd = Component("year_end")



from unittest.mock import patch

import pytest
from wbcore.contrib.authentication.factories import GroupFactory, UserFactory
from wbcore.contrib.authentication.models import Permission
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.directory.models import Person
from wbcore.contrib.workflow.factories import (
    FinishStepFactory,
    JoinStepFactory,
    UserStepFactory,
)
from wbcore.contrib.workflow.models import Condition, ProcessStep, Step
from wbcore.contrib.workflow.models.step import activate_step, process_can_finish
from wbcore.test.utils import get_or_create_superuser


@pytest.mark.django_db
class TestStep:
    @patch("wbcore.contrib.workflow.models.step.Step.run")
    def test_activate_step_no_status(self, mock_run, step_factory, process_factory):
        step = step_factory(status=None)
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.Step.run")
    def test_activate_step_no_instance(self, mock_run, step_factory, process_factory):
        step = step_factory()
        process = process_factory(workflow=step.workflow, instance=None, instance_id=None)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.Step.run")
    def test_activate_step_wrong_status_field(self, mock_run, step_factory, process_factory):
        step = step_factory(workflow__status_field="test")
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.Step.run")
    def test_activate_step_wrong_status_field_type(self, mock_run, step_factory, process_factory):
        step = step_factory(workflow__status_field="birthday")
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).birthday == process.instance.birthday
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.UserStep.run")
    def test_activate_user_step(self, mock_run, user_step_factory, process_factory):
        step = user_step_factory(permission=Permission.objects.last())
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            permission=step.permission,
            status=step.status,
            group=step.group,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.DecisionStep.run")
    def test_activate_decision_step(self, mock_run, decision_step_factory, process_factory):
        step = decision_step_factory()
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.SplitStep.run")
    def test_activate_split_step(self, mock_run, split_step_factory, process_factory):
        step = split_step_factory()
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.ScriptStep.run")
    def test_activate_script_step(self, mock_run, script_step_factory, process_factory):
        step = script_step_factory()
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.EmailStep.run")
    def test_activate_email_step(self, mock_run, email_step_factory, process_factory):
        step = email_step_factory()
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.FinishStep.run")
    def test_activate_finish_step(self, mock_run, finish_step_factory, process_factory):
        step = finish_step_factory()
        process = process_factory(workflow=step.workflow)
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.JoinStep.run")
    def test_activate_join_step(self, mock_run, process_step_factory, join_step_factory, process_factory):
        step = join_step_factory()
        process = process_factory(workflow=step.workflow)
        process_step_factory(
            state=ProcessStep.StepState.WAITING,
            step=step,
            status=step.status,
            process=process,
        )
        process_step_factory(
            state=ProcessStep.StepState.WAITING,
            step=step,
            status=step.status,
            finished=None,
        )
        process_step_factory()
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.JoinStep.run")
    def test_activate_step_too_many_waiting_steps(
        self, mock_run, process_step_factory, join_step_factory, process_factory
    ):
        step = join_step_factory()
        process = process_factory(workflow=step.workflow)
        process_step_factory(
            state=ProcessStep.StepState.WAITING,
            step=step,
            status=step.status,
            process=process,
            finished=None,
        )
        process_step_factory(
            state=ProcessStep.StepState.WAITING,
            step=step,
            status=step.status,
            process=process,
            finished=None,
        )
        process_step_factory()
        activate_step(step.pk, process.pk)
        process_step_qs = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        )
        assert process_step_qs.count() == 1
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert mock_run.call_args.args == (process_step_qs.first(),)

    @patch("wbcore.contrib.workflow.models.step.JoinStep.run")
    def test_activate_waiting_join_step(self, mock_run, process_step_factory, join_step_factory, process_factory):
        step = join_step_factory()
        process = process_factory(workflow=step.workflow)
        process_step_factory(
            state=ProcessStep.StepState.WAITING,
            step=step,
            status=step.status,
            process=process,
        )
        waiting_process_step = process_step_factory(
            state=ProcessStep.StepState.WAITING,
            step=step,
            status=step.status,
            process=process,
            finished=None,
        )
        process_step_factory()
        activate_step(step.pk, process.pk)
        assert not ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process=process,
            step=step,
            status=step.status,
        ).exists()
        assert Person.objects.get(id=process.instance.pk).first_name == step.status
        assert str(mock_run.call_args.args[0].pk) == waiting_process_step.pk

    def test_get_outgoing_transitions(self, random_child_step_factory, transition_factory):
        step = random_child_step_factory()
        transition = transition_factory(from_step=step)
        transition_factory()
        transition_factory(to_step=step)
        assert list(step.get_outgoing_transitions()) == [transition]

    def test_get_incoming_transitions(self, random_child_step_factory, transition_factory):
        step = random_child_step_factory()
        transition_factory(from_step=step)
        transition_factory()
        transition = transition_factory(to_step=step)
        assert list(step.get_incoming_transitions()) == [transition]

    def test_get_previous_steps(
        self,
        join_step_factory,
        split_step_factory,
        finish_step_factory,
        transition_factory,
    ):
        to_step = join_step_factory()
        from_step = split_step_factory()
        transition1 = transition_factory(to_step=to_step, from_step=from_step)
        transition2 = transition_factory(to_step=to_step, from_step=from_step)
        transition_factory(to_step=finish_step_factory(), from_step=to_step)
        transition_factory()
        previous_steps = to_step.get_previous_steps()
        previous_steps = [step.id for step in previous_steps]
        assert set(previous_steps) == {
            transition1.from_step.id,
            transition2.from_step.id,
        }

    def test_get_all_valid_outgoing_transitions(
        self,
        random_child_step_factory,
        transition_factory,
        condition_factory,
        process_step_factory,
    ):
        person = PersonFactory()
        step = random_child_step_factory()
        invalid_transition = transition_factory(from_step=step, to_step__workflow=step.workflow)
        condition_factory(
            transition=invalid_transition,
            attribute_name="first_name",
            expected_value=person.first_name,
            operator=Condition.Operator.EQ,
            negate_operator=True,
        )
        condition_factory(
            transition=invalid_transition,
            attribute_name="last_name",
            expected_value=person.last_name,
            operator=Condition.Operator.EQ,
            negate_operator=False,
        )
        valid_transition = transition_factory(from_step=step, to_step__workflow=step.workflow)
        condition_factory(
            transition=valid_transition,
            attribute_name="first_name",
            expected_value=person.first_name,
            operator=Condition.Operator.EQ,
            negate_operator=False,
        )
        valid_transition_without_condition = transition_factory(from_step=step, to_step__workflow=step.workflow)
        process_step = process_step_factory(step=step, process__instance=person)
        transitions = step.get_all_valid_outgoing_transitions(process_step)
        assert set(transitions) == {
            valid_transition,
            valid_transition_without_condition,
        }

    def test_get_assigned_group(self, random_child_step_factory):
        random_step = random_child_step_factory(exclude_factories=[UserStepFactory])
        assert random_step.get_assigned_group() is None

    def test_get_assigned_user(self, random_child_step_factory):
        random_step = random_child_step_factory(exclude_factories=[UserStepFactory])
        assert random_step.get_assigned_user() is None

    def test_get_casted_step(self, random_child_step_factory):
        random_step_child = random_child_step_factory()
        parent_step = random_step_child.step_ptr
        assert parent_step.get_casted_step().__class__ == random_step_child.__class__

    def test_user_can_see_step_user_permission(self, random_child_step_factory):
        step = random_child_step_factory(permission=Permission.objects.last())
        user = UserFactory()
        user.user_permissions.add(step.permission)
        assert step.user_can_see_step(user)

    def test_user_can_see_step_no_permission(self, random_child_step_factory):
        step = random_child_step_factory()
        user = UserFactory()
        assert step.user_can_see_step(user)

    def test_user_can_see_step_superuser(self, random_child_step_factory):
        step = random_child_step_factory(permission=Permission.objects.last())
        user = get_or_create_superuser()
        assert step.user_can_see_step(user)

    def test_user_can_see_step_group_permission(self, random_child_step_factory):
        step = random_child_step_factory(permission=Permission.objects.last())
        group = GroupFactory(permissions=[step.permission])
        user = UserFactory(groups=[group])
        assert step.user_can_see_step(user)

    def test_user_can_see_step_user_no_permission(self, random_child_step_factory):
        step = random_child_step_factory(permission=Permission.objects.last())
        user = UserFactory()
        assert not step.user_can_see_step(user)

    @patch("wbcore.contrib.workflow.models.step.Step.finish")
    def test_set_finished(self, mock_finish, process_step_factory, random_child_step_factory):
        step = random_child_step_factory(exclude_factories=[FinishStepFactory])
        process_step = process_step_factory(step=step)
        step.set_finished(process_step)
        assert process_step.state == ProcessStep.StepState.FINISHED
        assert mock_finish.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.FinishStep.finish")
    def test_set_finished_finish_step(self, mock_finish, process_step_factory, finish_step_factory):
        step = finish_step_factory()
        process_step = process_step_factory(step=step)
        step.set_finished(process_step)
        assert process_step.state == ProcessStep.StepState.FINISHED
        assert mock_finish.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.process_can_finish.delay")
    def test_set_failed(self, mock_can_finish, process_step_factory, random_child_step_factory):
        step = random_child_step_factory()
        process_step = process_step_factory(step=step, error_message=None)
        error_message = "Error message"
        step.set_failed(process_step, error_message)
        assert process_step.state == ProcessStep.StepState.FAILED
        assert process_step.error_message == error_message
        assert mock_can_finish.call_args.args == (process_step.process.pk,)

    @patch("wbcore.contrib.workflow.models.step.process_can_finish.delay")
    def test_set_failed_has_error_message(self, mock_can_finish, process_step_factory, random_child_step_factory):
        step = random_child_step_factory()
        error_message = "Error message"
        process_step = process_step_factory(step=step, error_message=error_message)
        step.set_failed(process_step, "Test")
        assert process_step.state == ProcessStep.StepState.FAILED
        assert process_step.error_message == error_message
        assert mock_can_finish.call_args.args == (process_step.process.pk,)

    @patch("wbcore.contrib.workflow.models.step.Step.finish")
    def test_set_canceled(self, mock_finish, process_step_factory, random_child_step_factory):
        step = random_child_step_factory(exclude_factories=[FinishStepFactory])
        process_step = process_step_factory(step=step)
        step.set_canceled(process_step)
        assert process_step.state == ProcessStep.StepState.CANCELED
        assert mock_finish.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.FinishStep.finish")
    def test_set_canceled_finish_step(self, mock_finish, process_step_factory, finish_step_factory):
        step = finish_step_factory()
        process_step = process_step_factory(step=step)
        step.set_canceled(process_step)
        assert process_step.state == ProcessStep.StepState.CANCELED
        assert mock_finish.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.Step.start_next_step")
    def test_execute_single_next_step(
        self,
        mock_next,
        process_step_factory,
        transition_factory,
        random_child_step_factory,
    ):
        step = random_child_step_factory()
        process_step = process_step_factory(step=step)
        transition = transition_factory(from_step=step)
        transition_factory()
        step.execute_single_next_step(process_step)
        assert mock_next.call_args.args == (process_step, transition)

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    def test_execute_single_next_step_no_transitions(
        self,
        mock_failed,
        process_step_factory,
        transition_factory,
        random_child_step_factory,
    ):
        step = random_child_step_factory(exclude_factories=[UserStepFactory])
        process_step = process_step_factory(step=step)
        transition_factory()
        step.execute_single_next_step(process_step)
        assert mock_failed.call_args.args[0] == process_step

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    def test_execute_single_next_step_multiple_transitions(
        self,
        mock_failed,
        process_step_factory,
        transition_factory,
        random_child_step_factory,
    ):
        step = random_child_step_factory(exclude_factories=[UserStepFactory])
        process_step = process_step_factory(step=step)
        transition_factory(from_step=step)
        transition_factory(from_step=step)
        step.execute_single_next_step(process_step)
        assert mock_failed.call_args.args[0] == process_step

    @patch("wbcore.contrib.workflow.models.step.Step.set_finished")
    def test_run(self, mock_finished, process_step_factory):
        process_step = process_step_factory()
        step = process_step.step
        step.run(process_step)
        assert mock_finished.call_args.args == (process_step,)

    @pytest.mark.parametrize(
        "state, expected",
        [
            ("Finished", True),
            ("Canceled", True),
            ("Active", False),
            ("Waiting", False),
            ("Failed", False),
        ],
    )
    def test_finish(self, state, expected, process_step_factory, random_child_step_factory):
        step = random_child_step_factory(exclude_factories=[FinishStepFactory])
        process_step = process_step_factory(step=step, state=state, finished=None)
        step.finish(process_step)
        assert bool(process_step.finished) is expected

    @patch("wbcore.contrib.workflow.models.step.activate_step.delay")
    @patch("wbcore.contrib.workflow.models.step.Step.set_finished")
    def test_start_next_step(self, mock_finish, mock_activate, transition_factory, process_step_factory):
        transition = transition_factory()
        process_step = process_step_factory()
        Step.start_next_step(process_step, transition)
        assert mock_finish.call_args.args == (process_step,)
        assert mock_activate.call_args.args == (
            transition.to_step.pk,
            process_step.process.pk,
        )

    @patch("wbcore.contrib.workflow.models.process.Process.set_failed")
    def test_process_can_finish_no_active(self, mock_failed, process_step_factory):
        process_step = process_step_factory(state=ProcessStep.StepState.CANCELED)
        process = process_step.process
        process_step_factory(state=ProcessStep.StepState.FINISHED, process=process)
        process_step_factory(state=ProcessStep.StepState.FAILED, process=process)
        process_step_factory(state=ProcessStep.StepState.ACTIVE)
        process_can_finish(process.pk)
        assert mock_failed.call_count == 1

    @patch("wbcore.contrib.workflow.models.process.Process.set_failed")
    def test_process_can_finish_no_join_steps(
        self,
        mock_failed,
        process_step_factory,
        random_child_step_factory,
        process_factory,
    ):
        step = random_child_step_factory(exclude_factories=[JoinStepFactory])
        process = process_factory(workflow=step.workflow)
        process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, process=process)
        process_can_finish(process.pk)
        assert not mock_failed.called

    # """
    # Test Workflow:
    #                          8
    #                        /   \                    # noqa: W605
    #              4 - 6 - 7       14
    #            /           \   /    \               # noqa: W605
    #           /              9        \             # noqa: W605
    # 1 - 2 - 3                            16 - 17
    #           \               12       /            # noqa: W605
    #            \             /  \    /              # noqa: W605
    #              5 - 10 - 11      15
    #                          \  /                   # noqa: W605
    #                           13
    # """

    @patch("wbcore.contrib.workflow.models.process.Process.set_failed")
    def test_process_can_finish_join_steps_can_be_reached(
        self,
        mock_failed,
        workflow_factory,
        split_step_factory,
        process_factory,
        process_step_factory,
        random_child_step_factory,
        start_step_factory,
        transition_factory,
        join_step_factory,
        finish_step_factory,
    ):
        workflow = workflow_factory()
        step1 = start_step_factory(workflow=workflow)
        step2 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step1, to_step=step2)
        step3 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step2, to_step=step3)
        step4 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step5 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step3, to_step=step4)
        transition_factory(from_step=step3, to_step=step5)
        step6 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step4, to_step=step6)
        step7 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step6, to_step=step7)
        step8 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step9 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step7, to_step=step8)
        transition_factory(from_step=step7, to_step=step9)
        step10 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step5, to_step=step10)
        step11 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step10, to_step=step11)
        step12 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step13 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step11, to_step=step12)
        transition_factory(from_step=step11, to_step=step13)
        step14 = join_step_factory(wait_for_all=True, workflow=workflow)
        transition_factory(from_step=step8, to_step=step14)
        transition_factory(from_step=step9, to_step=step14)
        step15 = join_step_factory(wait_for_all=False, workflow=workflow)
        transition_factory(from_step=step12, to_step=step15)
        transition_factory(from_step=step13, to_step=step15)
        step16 = join_step_factory(wait_for_all=True, workflow=workflow)
        transition_factory(from_step=step14, to_step=step16)
        transition_factory(from_step=step15, to_step=step16)
        step17 = finish_step_factory(workflow=workflow)
        transition_factory(from_step=step16, to_step=step17)

        process = process_factory(workflow=workflow)
        process_step_factory(step=step12, state=ProcessStep.StepState.FAILED, process=process)
        process_step_factory(step=step11, state=ProcessStep.StepState.ACTIVE, process=process)
        for step in [
            step1,
            step2,
            step3,
            step4,
            step5,
            step6,
            step7,
            step8,
            step9,
            step10,
            step14,
        ]:
            process_step_factory(step=step, state=ProcessStep.StepState.FINISHED, process=process)
        process_can_finish(process.pk)
        assert not mock_failed.called

    @patch("wbcore.contrib.workflow.models.process.Process.set_failed")
    def test_process_can_finish_waiting_join_step_one_failed(
        self,
        mock_failed,
        workflow_factory,
        split_step_factory,
        process_factory,
        process_step_factory,
        random_child_step_factory,
        start_step_factory,
        transition_factory,
        join_step_factory,
        finish_step_factory,
    ):
        workflow = workflow_factory()
        step1 = start_step_factory(workflow=workflow)
        step2 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step1, to_step=step2)
        step3 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step2, to_step=step3)
        step4 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step5 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step3, to_step=step4)
        transition_factory(from_step=step3, to_step=step5)
        step6 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step4, to_step=step6)
        step7 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step6, to_step=step7)
        step8 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step9 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step7, to_step=step8)
        transition_factory(from_step=step7, to_step=step9)
        step10 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step5, to_step=step10)
        step11 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step10, to_step=step11)
        step12 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step13 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step11, to_step=step12)
        transition_factory(from_step=step11, to_step=step13)
        step14 = join_step_factory(wait_for_all=True, workflow=workflow)
        transition_factory(from_step=step8, to_step=step14)
        transition_factory(from_step=step9, to_step=step14)
        step15 = join_step_factory(wait_for_all=False, workflow=workflow)
        transition_factory(from_step=step12, to_step=step15)
        transition_factory(from_step=step13, to_step=step15)
        step16 = join_step_factory(wait_for_all=True, workflow=workflow)
        transition_factory(from_step=step14, to_step=step16)
        transition_factory(from_step=step15, to_step=step16)
        step17 = finish_step_factory(workflow=workflow)
        transition_factory(from_step=step16, to_step=step17)

        process = process_factory(workflow=workflow)
        process_step_factory(step=step8, state=ProcessStep.StepState.FAILED, process=process)
        process_step_factory(step=step14, state=ProcessStep.StepState.WAITING, process=process)
        for step in [
            step1,
            step2,
            step3,
            step4,
            step5,
            step6,
            step7,
            step9,
            step10,
            step11,
            step12,
            step13,
            step15,
        ]:
            process_step_factory(step=step, state=ProcessStep.StepState.FINISHED, process=process)
        process_can_finish(process.pk)
        assert mock_failed.called

    @patch("wbcore.contrib.workflow.models.process.Process.set_failed")
    def test_process_can_finish_failed_at_beginning(
        self,
        mock_failed,
        workflow_factory,
        split_step_factory,
        process_factory,
        process_step_factory,
        random_child_step_factory,
        start_step_factory,
        transition_factory,
        join_step_factory,
        finish_step_factory,
    ):
        workflow = workflow_factory()
        step1 = start_step_factory(workflow=workflow)
        step2 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step1, to_step=step2)
        step3 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step2, to_step=step3)
        step4 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step5 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step3, to_step=step4)
        transition_factory(from_step=step3, to_step=step5)
        step6 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step4, to_step=step6)
        step7 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step6, to_step=step7)
        step8 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step9 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step7, to_step=step8)
        transition_factory(from_step=step7, to_step=step9)
        step10 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step5, to_step=step10)
        step11 = split_step_factory(workflow=workflow)
        transition_factory(from_step=step10, to_step=step11)
        step12 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        step13 = random_child_step_factory(exclude_factories=[JoinStepFactory], workflow=workflow)
        transition_factory(from_step=step11, to_step=step12)
        transition_factory(from_step=step11, to_step=step13)
        step14 = join_step_factory(wait_for_all=True, workflow=workflow)
        transition_factory(from_step=step8, to_step=step14)
        transition_factory(from_step=step9, to_step=step14)
        step15 = join_step_factory(wait_for_all=False, workflow=workflow)
        transition_factory(from_step=step12, to_step=step15)
        transition_factory(from_step=step13, to_step=step15)
        step16 = join_step_factory(wait_for_all=True, workflow=workflow)
        transition_factory(from_step=step14, to_step=step16)
        transition_factory(from_step=step15, to_step=step16)
        step17 = finish_step_factory(workflow=workflow)
        transition_factory(from_step=step16, to_step=step17)

        process = process_factory(workflow=workflow)
        process_step_factory(step=step6, state=ProcessStep.StepState.FAILED, process=process)
        process_step_factory(step=step15, state=ProcessStep.StepState.ACTIVE, process=process)
        for step in [
            step1,
            step2,
            step3,
            step4,
            step5,
            step10,
            step11,
            step12,
            step13,
            step15,
        ]:
            process_step_factory(step=step, state=ProcessStep.StepState.FINISHED, process=process)
        process_can_finish(process.pk)
        assert mock_failed.called



from typing import Iterable

from django.contrib.messages import get_messages
from django.contrib.staticfiles import finders
from django.db import IntegrityError
from django.db.models import F, Model, QuerySet
from django.db.models.expressions import OrderBy
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.templatetags.static import static
from django.utils.functional import cached_property
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter, SearchFilter
from rest_framework.mixins import CreateModelMixin as OriginalCreateModelMixin
from rest_framework.mixins import DestroyModelMixin as OriginalDestroyModelMixin
from rest_framework.mixins import ListModelMixin as OriginalListModelMixin
from rest_framework.mixins import RetrieveModelMixin as OriginalRetrieveModelMixin
from rest_framework.mixins import UpdateModelMixin as OriginalUpdateModelMixin
from rest_framework.pagination import BasePagination
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.status import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_405_METHOD_NOT_ALLOWED,
    HTTP_409_CONFLICT,
)

from wbcore.filters import DjangoFilterBackend
from wbcore.messages import InMemoryMessageStorage


def map_ordering(order_str: str) -> OrderBy:
    asc_desc = "desc" if order_str[0] == "-" else "asc"
    remove_strings = ["-", "__nulls_last", "__nulls_first"]
    clean_order_str = order_str
    for string in remove_strings:
        clean_order_str = clean_order_str.replace(string, "")
    asc_desc_function = getattr(F(clean_order_str), asc_desc)
    if "__nulls_last" in order_str:
        return asc_desc_function(nulls_last=True)
    elif "__nulls_first" in order_str:
        return asc_desc_function(nulls_first=True)
    return asc_desc_function()


class WBCoreOrderingFilter(OrderingFilter):
    def get_mandatory_ordering_fields(self, queryset, view) -> list[str]:
        """
        Given the queryset and the view, extract the mandatory field for the final ordering. Default to ["id"]

        To keep it simple, we assume that all view have a id field defined in the serializer. If this is not the case, we expect the dev to override this method and return an empty list.
        """
        if (fct := getattr(view, "get_mandatory_ordering_fields", None)) and callable(fct):
            fields = fct()
        else:
            fields = ["id"]
        return fields

    def filter_queryset(self, request, queryset, view):
        ordering = self.get_ordering(request, queryset, view)
        mandatory_ordering_fields = self.get_mandatory_ordering_fields(queryset, view)
        if ordering or mandatory_ordering_fields:
            return queryset.order_by(*map(map_ordering, ordering), *mandatory_ordering_fields)

        return queryset

    def get_valid_fields(self, queryset, view, context={}):
        valid_fields = view.get_ordering_fields()
        if valid_fields is None:
            # Default to allowing filtering on serializer fields
            return self.get_default_valid_fields(queryset, view, context)

        elif valid_fields == "__all__":
            # View explicitly allows filtering on any model field
            valid_fields = [(field.name, field.verbose_name) for field in queryset.model._meta.fields]
            valid_fields += [(key, key.title().split("__")) for key in queryset.query.annotations]
        else:
            valid_fields = [(item, item) if isinstance(item, str) else item for item in valid_fields]
        return valid_fields


class FilterMixin:
    filter_backends = (DjangoFilterBackend, SearchFilter, WBCoreOrderingFilter)
    filterset_fields = {}
    search_fields = []
    ordering_fields = ordering = ("id",)


class DocumentationMixin:
    def _get_documentation_url(self, detail):
        INSTANCE_DOCUMENTATION = getattr(self, "INSTANCE_DOCUMENTATION", None)
        LIST_DOCUMENTATION = getattr(self, "LIST_DOCUMENTATION", None)
        doc = INSTANCE_DOCUMENTATION if detail else LIST_DOCUMENTATION

        if doc and finders.find(doc):
            return static(doc)
        return None


class ActionMixin:
    kwargs: dict
    request: Request
    action: str

    def get_action(self) -> str:
        if self.action == "metadata":
            if bool(self.request.GET.get("new_mode")):
                return "create-metadata"
            if bool(self.kwargs.get("pk")):
                return "retrieve-metadata"
            else:
                return "list-metadata"
        return self.action


class MessageMixin:
    def add_messages(
        self,
        request: Request,
        queryset: QuerySet | None = None,
        paginated_queryset: BasePagination | None = None,
        instance: Model | None = None,
        initial: bool = False,
    ): ...

    def _get_messages(
        self,
        request: Request,
        queryset: QuerySet | None = None,
        paginated_queryset: BasePagination | None = None,
        instance: Model | None = None,
        initial: bool = False,
    ) -> Iterable[dict[str, str]]:
        # We call add_messages to collect messages from the view which weren't created at other places
        # because they might need some more information under which circumstances they are being called
        self.add_messages(
            request=request,
            queryset=queryset,
            paginated_queryset=paginated_queryset,
            instance=instance,
            initial=initial,
        )

        # We get the storage backend and yield all messages in a serialized form
        # This only works if the storage backend is our custom InMemoryMessageStorage
        # We inject request._request purely for typing, as request will defer attribute
        # lookup to its internal getattribute method. But the get_messages method only
        # accepts a HttpRequest object which is DRF request simply wraps
        storage = get_messages(request._request)
        if isinstance(storage, InMemoryMessageStorage):
            yield from storage.serialize_messages()


class ListModelMixin(MessageMixin, OriginalListModelMixin):
    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        if queryset.exists():
            queryset = self.filter_queryset(queryset)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            response = self.get_paginated_response(serializer.data)
            response.data["messages"] = list(
                self._get_messages(
                    request, queryset=queryset, paginated_queryset=page, initial=self._paginator.is_initial()
                )
            )

            if hasattr(self, "get_aggregates"):
                response.data["aggregates"] = self.get_aggregates(queryset=queryset, paginated_queryset=page)
        else:
            serializer = self.get_serializer(queryset, many=True)
            response = Response(serializer.data)
            response.data = {"results": response.data}

            response.data["messages"] = list(self._get_messages(request, queryset=queryset, initial=True))

            if hasattr(self, "get_aggregates"):
                response.data["aggregates"] = self.get_aggregates(queryset=queryset, paginated_queryset=queryset)

        return response


class RetrieveModelMixin(MessageMixin, OriginalRetrieveModelMixin):
    def retrieve(self, request, *args, **kwargs):
        response = super().retrieve(request, *args, **kwargs)
        response.data = {"instance": response.data}
        try:
            instance = self.get_object()
        except Http404:
            instance = None
        response.data["messages"] = list(self._get_messages(request, instance=instance))
        return response


class CreateModelMixin(OriginalCreateModelMixin):
    def create(self, request, *args, **kwargs):
        # If not create endpoint is defined then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=False)._get_create_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        response = super().create(request, *args, **kwargs)
        response.data = {"instance": response.data, "messages": list(self._get_messages(request))}
        return response


class UpdateModelMixin(OriginalUpdateModelMixin):
    def update(self, request, *args, **kwargs):
        # If no instance endpoint is defined, then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=True)._get_update_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        response = super().update(request, *args, **kwargs)
        response.data = {"instance": response.data}
        try:
            instance = self.get_object()
        except Http404:
            instance = None
        response.data["messages"] = list(self._get_messages(request, instance=instance))
        return response


class DestroyModelMixin(OriginalDestroyModelMixin):
    def destroy(self, request, *args, **kwargs):
        # If no delete endpoint is defined, then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=True)._get_delete_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        return super().destroy(request, *args, **kwargs)


class DestroyMultipleModelMixin:
    def destroy_multiple(self, request, *args, **kwargs):
        # If no delete endpoint is defined, then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=False)._get_delete_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        model = self.get_serializer_class().Meta.model
        app_label = model._meta.app_label

        queryset = model.objects.filter(id__in=request.data)
        destroyed = self.perform_destroy_multiple(queryset)
        return Response({"count": destroyed[1].get(f"{app_label}.{model.__name__}", 0)}, status=HTTP_204_NO_CONTENT)

    def perform_destroy_multiple(self, queryset):
        return queryset.delete()


class OrderableMixin:
    # This mixin will not work if the viewset model does not inherit from ordered_model.models.OrderedModel

    ordering = ("order",)

    @action(methods=["PATCH"], detail=True)
    def reorder(self, request: Request, pk=None, **kwargs) -> Response:
        instance = self.get_object()
        order = request.data.get("order")
        if order is None:
            return Response("No order received", status=HTTP_400_BAD_REQUEST)
        instance.to(order)
        return Response("Reordering Successful", status=HTTP_200_OK)


class ReparentMixin:
    PARENT_FIELD: str
    PARENT_MODEL: type[Model]

    @property
    def parent_field(self) -> str:
        return self.PARENT_FIELD

    @cached_property
    def model(self):
        return self.queryset.model

    @action(methods=["PATCH"], detail=True)
    def reparent(self, request: Request, pk=None, **kwargs) -> Response:
        instance = self.get_object()
        new_parent = get_object_or_404(self.PARENT_MODEL, pk=request.data.get("new_parent"))
        old_parent = getattr(instance, self.parent_field)
        if old_parent != new_parent:
            setattr(instance, self.parent_field, new_parent)
            try:
                instance.save()
            except IntegrityError:
                return Response("Cannot reparent due to conflict", status=HTTP_409_CONFLICT)
        return Response("Reparenting Successful", status=HTTP_200_OK)



import io
import logging
import re
from datetime import datetime
from io import BytesIO
from typing import Generator, Optional

import fabric
from django.conf import settings
from paramiko.ssh_exception import SSHException
from wbcore.contrib.io.backends.abstract import AbstractDataBackend
from wbcore.contrib.io.backends.utils import register
from wbcore.contrib.io.models import ImportCredential

logger = logging.getLogger("io")


@register("SFTP", save_data_in_import_source=True, passive_only=True)
class DataBackend(AbstractDataBackend):
    def __init__(
        self,
        import_credential: Optional[ImportCredential] = None,
        **kwargs,
    ):
        if not import_credential:
            raise ValueError("The sftp backend needs an import credential")

        self.username = import_credential.username
        self.password = import_credential.password
        self.host = import_credential.additional_resources.get("host", "")
        self.port = import_credential.additional_resources.get("port", "")
        super().__init__(**kwargs)

    def get_files(
        self,
        execution_time: datetime,
        sftp_folder,
        file_name_regex: Optional[str] = None,
        cleanup_files: Optional[bool] = getattr(settings, "WBIMPORT_EXPORT_SFTPBACKEND_CLEAN_FILES", False),
        **kwargs,
    ) -> Generator[tuple[str, BytesIO], None, None]:
        # Create Connection to the SFTP Server
        with fabric.Connection(
            host=self.host, port=self.port, user=self.username, connect_kwargs={"password": self.password}
        ) as conn:
            try:
                sftp = conn.sftp()
                # Change working directory
                sftp.chdir(sftp_folder)

                # Filter all the files we need
                file_names = sftp.listdir()
                if file_name_regex:
                    # Compile the regex for later filtering the files
                    file_names = filter(lambda x: re.match(file_name_regex, x), file_names)

                for file_name in file_names:
                    # Create a Buffer where we write the file to
                    sftp_file = io.BytesIO()
                    sftp.getfo(file_name, sftp_file)
                    yield file_name, sftp_file

                    if cleanup_files:
                        # Delete the file from the server
                        sftp.remove(file_name)
            except SSHException as e:
                logger.warning(
                    f"While fetching file from data backend {self.data_backend}, we encountered a SSH Exception: {e}"
                )



from rest_framework.reverse import reverse

from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ImportSourceModelViewSetEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcore:io:importsource-list", args=[], request=self.request)


class SourceModelViewSetEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcore:io:source-list", args=[], request=self.request)



from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import IntegerPreference, StringPreference
from dynamic_preferences.users.registries import user_preferences_registry

from wbcore.dynamic_preference.types import ChoicePreference, LanguageChoicePreference

wbcore = Section("wbcore")


@global_preferences_registry.register
class RetentionPeriod(IntegerPreference):
    section = wbcore
    name = "retention_period"
    default = 365

    verbose_name = _("Retention Period in Days")
    help_text = _(
        "When an object cannot be deleted and is disabled instead, it gets hidden from the queryset but not deleted. For compliance reasons we enable the retention for a specific period of days (defaults to a year)"
    )


@global_preferences_registry.register
class SystemUserEmailPeriod(StringPreference):
    section = wbcore
    name = "system_user_email"

    default = "system@stainly-bench.com"

    verbose_name = _("System User Email")
    help_text = _("System User Email")


@user_preferences_registry.register
class LanguagePreference(LanguageChoicePreference):
    choices = [
        ("de", "Deutsch"),
        ("fr", "Français"),
        ("en", "English"),
    ]

    section = wbcore
    name = "language"
    default = "en"

    verbose_name = _("System Language")
    help_text = _("System Language")


@user_preferences_registry.register
class DateFormatPreference(ChoicePreference):
    choices = [
        ("DD.MM.YYYY", "13.04.2007"),
        ("DD/MM/YYYY", "13/04/2007"),
        ("DD-MM-YYYY", "13-04-2007"),
        ("MM/DD/YYYY", "04/13/2007"),
        ("MM-DD-YYYY", "04-13-2007"),
        ("YYYY/MM/DD", "2007/04/13"),
        ("YYYY-MM-DD", "2007-04-13"),
        ("MMM DD, YYYY", "April 13, 2007"),
        ("DD MMM YYYY", "13 April 2007"),
        ("dddd, DD MMM YYYY", "Friday, 13 April 2007"),
        ("ddd, DD MMM YYYY", "Fri, 13 April 2007"),
        ("ddd Do MMMM YYYY", "Fri 13th April 2007"),
    ]

    section = wbcore
    name = "date_format"
    default = "YYYY-MM-DD"

    verbose_name = _("Date Format")
    help_text = _("Date Format")


@user_preferences_registry.register
class TimeFormatPreference(ChoicePreference):
    choices = [
        ("HH:mm", "14:05"),
        ("hh:mm", "02:05"),
        ("h:mm", "2:05"),
        ("HH [h] mm", "14 h 05"),
        ("HH[h]mm", "14h05"),
        ("h:mm A", "2:05 PM"),
        ("h:mm a", "2:05 pm"),
        ("hh:mm A", "02:05 PM"),
        ("hh:mm a", "02:05 pm"),
    ]

    section = wbcore
    name = "time_format"
    default = "HH:mm"

    verbose_name = _("Time Format")
    help_text = _("Time Format")



from django.utils.translation import gettext_lazy as _

from wbcore import filters as wb_filters
from wbcore.contrib.workflow.models import Data, Workflow


class DataFilter(wb_filters.FilterSet):
    workflow = wb_filters.ModelMultipleChoiceFilter(
        label=_("Workflows"),
        queryset=Workflow.objects.all(),
        endpoint=Workflow.get_representation_endpoint(),
        value_key=Workflow.get_representation_value_key(),
        label_key=Workflow.get_representation_label_key(),
    )

    data_type = wb_filters.MultipleChoiceFilter(choices=Data.DataType.choices, label=_("Data Types"))

    class Meta:
        model = Data
        fields = {
            "label": ["exact", "icontains"],
            "help_text": ["exact", "icontains"],
            "default": ["exact", "icontains"],
            "required": ["exact"],
        }



from .condition import ConditionEndpointConfig
from .data import DataEndpointConfig
from .process import ProcessEndpointConfig, AssignedProcessStepEndpointConfig
from .step import DisplayEndpointConfig, StepEndpointConfig
from .transition import TransitionEndpointConfig



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class PivotUserObjectPermissionEndpointViewConfig(EndpointViewConfig):
    def get_update_endpoint(self, **kwargs):
        return reverse(
            "wbcore:guardian:pivoteduserobjectpermission-list",
            args=[self.view.kwargs.get("content_type_id", None), self.view.kwargs.get("object_pk", None)],
            request=self.request,
        )

    def get_create_endpoint(self, **kwargs):
        return reverse(
            "wbcore:guardian:pivoteduserobjectpermission-list",
            args=[self.view.kwargs.get("content_type_id", None), self.view.kwargs.get("object_pk", None)],
            request=self.request,
        )



import factory

from wbcore.contrib.example_app.models import Stadium


class StadiumFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=16)
    standing_capacity = factory.Faker("pyint", min_value=100, max_value=100000)
    seating_capacity = factory.Faker("pyint", min_value=100, max_value=100000)
    # city = factory.SubFactory("wbcore.contrib.geography.factories.CityFactory")

    class Meta:
        model = Stadium



from typing import Optional

from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class UserPermissionModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="codename", label=_("Code Name")),
                dp.Field(key="content_type_repr", label=_("Content Type")),
            ]
        )


class UserProfileModelDisplay(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(3, "profile_image"), "prefix", "first_name", "last_name"],
                [repeat_field(3, "profile_image"), repeat_field(2, "email"), "birthday"],
                [repeat_field(3, "profile_image"), repeat_field(2, "profile"), "username"],
                [repeat_field(6, "token_section")],
            ],
            [
                create_simple_section(
                    "token_section",
                    _("Token (Sensitive information: handle with care)"),
                    [["generic_auth_token_key"], ["calendar_subscription_link"]],
                )
            ],
        )


class UserModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="email", label=_("Email")),
                dp.Field(key="is_active", label=_("Is Active")),
                dp.Field(key="date_joined", label=_("Joined")),
                dp.Field(key="last_connection", label=_("Latest Connection")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["username", "email"],
                [repeat_field(2, "profile")],
                [repeat_field(2, "is_active")],
                [repeat_field(2, "groups_section")],
                [repeat_field(2, "permissions_section")],
                [repeat_field(2, "user_activity_section")],
            ],
            [
                create_simple_section("groups_section", _("Groups"), [["groups"]], collapsed=True),
                create_simple_section("permissions_section", _("Permissions"), [["user_permissions"]], collapsed=True),
                create_simple_section(
                    "user_activity_section", _("User Activity"), [["user_activity"]], "user_activity", collapsed=True
                ),
            ],
        )



from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.enums import Unit
from wbcore.metadata.configs.display import Field, Legend, LegendItem, ListDisplay
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class NotificationTypeSettingDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return ListDisplay(
            fields=[
                Field(key="locked_icon", label=" ", width=Unit.PIXEL(40)),
                Field(key="notification_type", label="Notification", width=Unit.PIXEL(250)),
                Field(key="help_text", label="Help Text", width=Unit.PIXEL(500)),
                Field(key="enable_web", label="Web", width=Unit.PIXEL(100)),
                Field(key="enable_mobile", label="Mobile", width=Unit.PIXEL(100)),
                Field(key="enable_email", label="E-Mail", width=Unit.PIXEL(100)),
            ],
            legends=[
                Legend(
                    items=[
                        LegendItem(
                            icon=WBIcon.LOCK.icon,
                            label="Locked",
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["notification_type", "notification_type", "notification_type"],
                ["enable_web", "enable_mobile", "enable_email"],
            ]
        )


class NotificationTypeSettingEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbcore:notifications:notification_type_setting-list", args=[], request=self.request)

    def get_update_endpoint(self):
        return "{{_update_url}}"



import pytest
from rest_framework.exceptions import ValidationError

from wbcore.serializers import CharField, StringRelatedField, TextField
from wbcore.serializers.fields.types import WBCoreType


class TestCharField:
    def setup_method(self):
        self.field = CharField()
        self.field_with_placeholder = CharField(placeholder="Enter Text here")

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize("input, expected", [(1, "1"), ("a", "a")])
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    def test_to_internal_value_validation_error(self):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(None)

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.TEXT.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "decorators": [],
                "depends_on": [],
            },
        )

    def test_string_placeholder(self):
        assert self.field_with_placeholder.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "decorators": [],
                "depends_on": [],
                "placeholder": "Enter Text here",
            },
        )


@pytest.mark.django_db
class TestStringRelatedField:
    def setup_method(self):
        self.field = StringRelatedField()

    def test_not_none(self):
        assert self.field is not None

    # def test_to_representation(self, model_test):
    #     assert self.field.to_representation(model_test) == str(model_test)

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.TEXT.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": False,
                "read_only": True,
                "decorators": [],
                "depends_on": [],
            },
        )


class TestTextField:
    def setup_method(self):
        self.field = TextField()

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize("input, expected", [(123, "123"), ("abc", "abc")])
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    def test_to_internal_value_validation_error(self):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(None)

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.TEXTEDITOR.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "content_type": self.field.texteditor_content_type,
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "plugin_configs": {},
                "read_only": False,
                "decorators": [],
                "depends_on": [],
            },
        )



from enum import Enum


class ButtonDefaultColor(Enum):
    SUCCESS = "success"
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    PRIMARY = "primary"
    SECONDARY = "secondary"
    INHERIT = "inherit"  # 'inherit' means that the button will take the color of a surrounding button group


class ButtonType(Enum):
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"


class HyperlinkTarget(Enum):
    BLANK = "_blank"
    SELF = "_self"
    PARENT = "_parent"
    TOP = "_top"



from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display import create_simple_display
from wbcore.serializers import Serializer


class DocumentTypeButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(
                key="list_of_children",
                label=_("All Children"),
                icon=WBIcon.ENUMERATION.icon,
            ),
            bt.WidgetButton(
                key="list_of_documents",
                label=_("All Documents"),
                icon=WBIcon.DOCUMENT.icon,
            ),
        }


class DocumentButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        class SendMailSerializer(Serializer):
            to_email = wb_serializers.CharField(default=self.request.user.email, label=gettext_lazy("Recipient"))

            class Meta:
                fields = "to_email"

        return {
            bt.HyperlinkButton(key="download_file", label=_("Download"), icon=WBIcon.LINK.icon),
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcore:documents:document",),
                key="send_mail",
                icon=WBIcon.SEND.icon,
                label=_("Send Mail"),
                description_fields=_("""<p>Send this document as mail?</p>"""),
                action_label=_("Sending Mail"),
                title=_("Send document as mail"),
                serializer=SendMailSerializer,
                instance_display=create_simple_display([["to_email"]]),
            ),
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from .choices import ChoiceFilter, MultipleChoiceField, MultipleChoiceFilter
from .datetime import (
    TimeFilter,
    DateTimeFilter,
    DateFilter,
    DateRangeFilter,
    DateTimeRangeFilter,
    FinancialPerformanceDateRangeFilter,
)
from .models import ModelChoiceFilter, ModelMultipleChoiceFilter
from .numbers import NumberFilter, YearFilter, RangeSelectFilter
from .text import CharFilter
from .booleans import BooleanFilter
from .content_type import MultipleChoiceContentTypeFilter
from .multiple_lookups import MultipleLookupFilter



# Generated by Django 5.0.9 on 2024-11-11 12:00

import colorfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("example_app", "0006_league_season_period_player_is_active_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="player",
            options={"ordering": ("order",), "verbose_name": "Player", "verbose_name_plural": "Players"},
        ),
        migrations.AlterModelOptions(
            name="team",
            options={"ordering": ("order",), "verbose_name": "Team", "verbose_name_plural": "Teams"},
        ),
        migrations.AddField(
            model_name="eventtype",
            name="color",
            field=colorfield.fields.ColorField(
                blank=True, default=None, image_field=None, max_length=25, null=True, samples=None
            ),
        ),
        migrations.AddField(
            model_name="player",
            name="order",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False, verbose_name="order"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="team",
            name="order",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False, verbose_name="order"),
            preserve_default=False,
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, Exists, F, Model, OuterRef, Q, QuerySet
from django.utils.functional import cached_property
from guardian.shortcuts import assign_perm, remove_perm
from rest_framework.exceptions import ValidationError
from wbcore import serializers, viewsets
from wbcore.contrib.authentication.models.users import Permission, User
from wbcore.contrib.authentication.serializers import UserRepresentationSerializer
from wbcore.contrib.guardian.models import UserObjectPermission
from wbcore.contrib.guardian.viewsets.configs import (
    PivotUserObjectPermissionDisplayViewConfig,
    PivotUserObjectPermissionEndpointViewConfig,
    PivotUserObjectPermissionTitleViewConfig,
)


class PivotUserObjectPermissionModelViewSet(viewsets.ModelViewSet):
    queryset = UserObjectPermission.objects.all()

    display_config_class = PivotUserObjectPermissionDisplayViewConfig
    endpoint_config_class = PivotUserObjectPermissionEndpointViewConfig
    title_config_class = PivotUserObjectPermissionTitleViewConfig

    lookup_field = "user_id"

    @cached_property
    def permissions(self):
        return Permission.objects.filter(
            Q(content_type=self.kwargs.get("content_type_id"))
            & ~Q(codename__icontains="administrate")
            & ~Q(codename__icontains="import")
            & ~Q(codename__icontains="export")
        )

    @cached_property
    def linked_object(self):
        return ContentType.objects.get(id=self.kwargs.get("content_type_id")).get_object_for_this_type(
            id=self.kwargs.get("object_pk")
        )

    def get_queryset(self):
        queryset = super().get_queryset()

        if (content_type_id := self.kwargs.get("content_type_id")) and (object_pk := self.kwargs.get("object_pk")):
            queryset = queryset.filter(content_type_id=content_type_id, object_pk=object_pk)

        queryset = queryset.values("user").annotate(
            id=F("user__id"),
            userref=F("user__id"),
            count=Count("id"),  # This is needed as we need at least 1 aggregation function to have a group by
            non_editable=Exists(queryset.filter(user=OuterRef("user"), editable=False)),
            **{
                permission.codename: Exists(queryset.filter(user=OuterRef("user"), permission_id=permission.pk))
                for permission in self.permissions
            },
        )

        return queryset

    def get_serializer_class(self):
        def assign_and_remove_perms(
            self,
            perms: dict[str, bool],
            permissions: QuerySet[Permission],
            user: User,
            linked_object: Model,
            remove: bool = True,
        ):
            for codename, perm in perms.items():
                if UserObjectPermission.objects.filter(
                    user=user,
                    content_type_id=self.context["view"].kwargs.get("content_type_id"),
                    object_pk=linked_object.pk,
                    editable=False,
                ).exists():
                    raise ValidationError(
                        {
                            "non_field_errors": "This user permission is not editable as it was created through the system"
                        }
                    )
                if perm:
                    assign_perm(
                        permissions.get(codename=codename),
                        user,
                        linked_object,
                    )
                elif remove:
                    remove_perm(
                        permissions.get(codename=codename),
                        user,
                        linked_object,
                    )

        def create(self, validated_data):
            user = validated_data.pop("userref")
            view = self.context["view"]
            self.assign_and_remove_perms(validated_data, view.permissions, user, view.linked_object, remove=False)
            self.context["view"].kwargs["user_id"] = user.id
            return self.context["view"].get_object()

        def update(self, instance, validated_data):
            view = self.context["view"]
            self.assign_and_remove_perms(
                validated_data, view.permissions, User.objects.get(id=instance.get("user")), view.linked_object
            )
            return instance

        attributes = {
            "id": serializers.PrimaryKeyField(),
            "count": serializers.IntegerField(required=False, read_only=True),
            "userref": serializers.PrimaryKeyRelatedField(queryset=User.objects.all(), label="User"),
            "_userref": UserRepresentationSerializer(source="userref"),
            "non_editable": serializers.BooleanField(default=False),
            **{permission.codename: serializers.BooleanField(default=False) for permission in self.permissions},
            "Meta": type(
                "Meta",
                (),
                {
                    "model": User,
                    "fields": (
                        "id",
                        "count",
                        "userref",
                        "_userref",
                        "non_editable",
                        *[perm.codename for perm in self.permissions],
                    ),
                },
            ),
            "update": update,
            "create": create,
            "assign_and_remove_perms": assign_and_remove_perms,
        }

        return type(
            "PivotedUserObjectPermissionModelSerializer",
            (serializers.ModelSerializer,),
            attributes,
        )
